@@ -35,7 +35,7 @@ import javax.servlet.ServletResponse;
 /**
  * Provides an abstract class to be subclassed to create
  * an HTTP servlet suitable for a Web site. A subclass of
- * <code>HttpServlet</code> must override at least 
+ * <code>HttpServlet</code> must override at least
  * one method, usually one of these:
  *
  * <ul>
@@ -43,27 +43,27 @@ import javax.servlet.ServletResponse;
  * <li> <code>doPost</code>, for HTTP POST requests
  * <li> <code>doPut</code>, for HTTP PUT requests
  * <li> <code>doDelete</code>, for HTTP DELETE requests
- * <li> <code>init</code> and <code>destroy</code>, 
+ * <li> <code>init</code> and <code>destroy</code>,
  * to manage resources that are held for the life of the servlet
  * <li> <code>getServletInfo</code>, which the servlet uses to
- * provide information about itself 
+ * provide information about itself
  * </ul>
  *
  * <p>There's almost no reason to override the <code>service</code>
  * method. <code>service</code> handles standard HTTP
  * requests by dispatching them to the handler methods
- * for each HTTP request type (the <code>do</code><i>XXX</i>
+ * for each HTTP request type (the <code>do</code><i>Method</i>
  * methods listed above).
  *
- * <p>Likewise, there's almost no reason to override the 
+ * <p>Likewise, there's almost no reason to override the
  * <code>doOptions</code> and <code>doTrace</code> methods.
- * 
+ *
  * <p>Servlets typically run on multithreaded servers,
  * so be aware that a servlet must handle concurrent
  * requests and be careful to synchronize access to shared resources.
  * Shared resources include in-memory data such as
  * instance or class variables and external objects
- * such as files, database connections, and network 
+ * such as files, database connections, and network
  * connections.
  * See the
  * <a href="http://java.sun.com/Series/Tutorial/java/threads/multithreaded.html">
@@ -87,24 +87,24 @@ public abstract class HttpServlet extends GenericServlet {
 
     private static final String HEADER_IFMODSINCE = "If-Modified-Since";
     private static final String HEADER_LASTMOD = "Last-Modified";
-    
+
     private static final String LSTRING_FILE =
         "javax.servlet.http.LocalStrings";
     private static ResourceBundle lStrings =
         ResourceBundle.getBundle(LSTRING_FILE);
-   
-    
+
+
     /**
      * Does nothing, because this is an abstract class.
      */
     public HttpServlet() {
         // NOOP
     }
-    
-    
+
+
     /**
      * Called by the server (via the <code>service</code> method) to
-     * allow a servlet to handle a GET request. 
+     * allow a servlet to handle a GET request.
      *
      * <p>Overriding this method to support a GET request also
      * automatically supports an HTTP HEAD request. A HEAD
@@ -112,7 +112,7 @@ public abstract class HttpServlet extends GenericServlet {
      * response, only the request header fields.
      *
      * <p>When overriding this method, read the request data,
-     * write the response headers, get the response's writer or 
+     * write the response headers, get the response's writer or
      * output stream object, and finally, write the response data.
      * It's best to include content type and encoding. When using
      * a <code>PrintWriter</code> object to return the response,
@@ -125,7 +125,7 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * <p>Where possible, set the Content-Length header (with the
      * {@link javax.servlet.ServletResponse#setContentLength} method),
-     * to allow the servlet container to use a persistent connection 
+     * to allow the servlet container to use a persistent connection
      * to return its response to the client, improving performance.
      * The content length is automatically set if the entire response fits
      * inside the response buffer.
@@ -141,10 +141,10 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * <p>The GET method should also be idempotent, meaning
      * that it can be safely repeated. Sometimes making a
-     * method safe also makes it idempotent. For example, 
+     * method safe also makes it idempotent. For example,
      * repeating queries is both safe and idempotent, but
      * buying a product online or modifying data is neither
-     * safe nor idempotent. 
+     * safe nor idempotent.
      *
      * <p>If the request is incorrectly formatted, <code>doGet</code>
      * returns an HTTP "Bad Request" message.
@@ -156,14 +156,14 @@ public abstract class HttpServlet extends GenericServlet {
      * @param resp  an {@link HttpServletResponse} object that
      *                  contains the response the servlet sends
      *                  to the client
-     * 
-     * @exception IOException   if an input or output error is 
+     *
+     * @exception IOException   if an input or output error is
      *                              detected when the servlet handles
      *                              the GET request
      *
      * @exception ServletException  if the request for the GET
      *                                  could not be handled
-     * 
+     *
      * @see javax.servlet.ServletResponse#setContentType
      */
     protected void doGet(HttpServletRequest req, HttpServletResponse resp)
@@ -191,7 +191,7 @@ public abstract class HttpServlet extends GenericServlet {
      * This makes browser and proxy caches work more effectively,
      * reducing the load on server and network resources.
      *
-     * @param req   the <code>HttpServletRequest</code> 
+     * @param req   the <code>HttpServletRequest</code>
      *                  object that is sent to the servlet
      *
      * @return  a <code>long</code> integer specifying
@@ -257,28 +257,28 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * <p>When overriding this method, read the request data,
      * write the response headers, get the response's writer or output
-     * stream object, and finally, write the response data. It's best 
+     * stream object, and finally, write the response data. It's best
      * to include content type and encoding. When using a
-     * <code>PrintWriter</code> object to return the response, set the 
-     * content type before accessing the <code>PrintWriter</code> object. 
+     * <code>PrintWriter</code> object to return the response, set the
+     * content type before accessing the <code>PrintWriter</code> object.
      *
      * <p>The servlet container must write the headers before committing the
-     * response, because in HTTP the headers must be sent before the 
+     * response, because in HTTP the headers must be sent before the
      * response body.
      *
      * <p>Where possible, set the Content-Length header (with the
      * {@link javax.servlet.ServletResponse#setContentLength} method),
-     * to allow the servlet container to use a persistent connection 
+     * to allow the servlet container to use a persistent connection
      * to return its response to the client, improving performance.
      * The content length is automatically set if the entire response fits
-     * inside the response buffer.  
+     * inside the response buffer.
      *
      * <p>When using HTTP 1.1 chunked encoding (which means that the response
-     * has a Transfer-Encoding header), do not set the Content-Length header. 
+     * has a Transfer-Encoding header), do not set the Content-Length header.
      *
      * <p>This method does not need to be either safe or idempotent.
      * Operations requested through POST can have side effects for
-     * which the user can be held accountable, for example, 
+     * which the user can be held accountable, for example,
      * updating stored data or buying items online.
      *
      * <p>If the HTTP POST request is incorrectly formatted,
@@ -292,8 +292,8 @@ public abstract class HttpServlet extends GenericServlet {
      * @param resp  an {@link HttpServletResponse} object that
      *                  contains the response the servlet sends
      *                  to the client
-     * 
-     * @exception IOException   if an input or output error is 
+     *
+     * @exception IOException   if an input or output error is
      *                              detected when the servlet handles
      *                              the request
      *
@@ -320,8 +320,8 @@ public abstract class HttpServlet extends GenericServlet {
      * Called by the server (via the <code>service</code> method)
      * to allow a servlet to handle a PUT request.
      *
-     * The PUT operation allows a client to 
-     * place a file on the server and is similar to 
+     * The PUT operation allows a client to
+     * place a file on the server and is similar to
      * sending a file by FTP.
      *
      * <p>When overriding this method, leave intact
@@ -377,7 +377,7 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * The DELETE operation allows a client to remove a document
      * or Web page from the server.
-     * 
+     *
      * <p>This method does not need to be either safe
      * or idempotent. Operations requested through
      * DELETE can have side effects for which users
@@ -396,7 +396,7 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * @param resp  the {@link HttpServletResponse} object that
      *                  contains the response the servlet returns
-     *                  to the client                                
+     *                  to the client
      *
      * @exception IOException   if an input or output error occurs
      *                              while the servlet is handling the
@@ -417,7 +417,7 @@ public abstract class HttpServlet extends GenericServlet {
             resp.sendError(HttpServletResponse.SC_BAD_REQUEST, msg);
         }
     }
-    
+
 
     private static Method[] getAllDeclaredMethods(Class<?> c) {
 
@@ -427,7 +427,7 @@ public abstract class HttpServlet extends GenericServlet {
 
         Method[] parentMethods = getAllDeclaredMethods(c.getSuperclass());
         Method[] thisMethods = c.getDeclaredMethods();
-        
+
         if ((parentMethods != null) && (parentMethods.length > 0)) {
             Method[] allMethods =
                 new Method[parentMethods.length + thisMethods.length];
@@ -447,7 +447,7 @@ public abstract class HttpServlet extends GenericServlet {
      * Called by the server (via the <code>service</code> method)
      * to allow a servlet to handle a OPTIONS request.
      *
-     * The OPTIONS request determines which HTTP methods 
+     * The OPTIONS request determines which HTTP methods
      * the server supports and
      * returns an appropriate header. For example, if a servlet
      * overrides <code>doGet</code>, this method returns the
@@ -456,7 +456,7 @@ public abstract class HttpServlet extends GenericServlet {
      * <p><code>Allow: GET, HEAD, TRACE, OPTIONS</code>
      *
      * <p>There's no need to override this method unless the
-     * servlet implements new HTTP methods, beyond those 
+     * servlet implements new HTTP methods, beyond those
      * implemented by HTTP 1.1.
      *
      * @param req   the {@link HttpServletRequest} object that
@@ -465,7 +465,7 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * @param resp  the {@link HttpServletResponse} object that
      *                  contains the response the servlet returns
-     *                  to the client                                
+     *                  to the client
      *
      * @exception IOException   if an input or output error occurs
      *                              while the servlet is handling the
@@ -479,7 +479,7 @@ public abstract class HttpServlet extends GenericServlet {
         throws ServletException, IOException {
 
         Method[] methods = getAllDeclaredMethods(this.getClass());
-        
+
         boolean ALLOW_GET = false;
         boolean ALLOW_HEAD = false;
         boolean ALLOW_POST = false;
@@ -487,22 +487,22 @@ public abstract class HttpServlet extends GenericServlet {
         boolean ALLOW_DELETE = false;
         boolean ALLOW_TRACE = true;
         boolean ALLOW_OPTIONS = true;
-        
+
         for (int i=0; i<methods.length; i++) {
             Method m = methods[i];
-            
+
             if (m.getName().equals("doGet")) {
                 ALLOW_GET = true;
                 ALLOW_HEAD = true;
             }
-            if (m.getName().equals("doPost")) 
+            if (m.getName().equals("doPost"))
                 ALLOW_POST = true;
             if (m.getName().equals("doPut"))
                 ALLOW_PUT = true;
             if (m.getName().equals("doDelete"))
                 ALLOW_DELETE = true;
         }
-        
+
         String allow = null;
         if (ALLOW_GET)
             allow=METHOD_GET;
@@ -524,18 +524,18 @@ public abstract class HttpServlet extends GenericServlet {
         if (ALLOW_OPTIONS)
             if (allow==null) allow=METHOD_OPTIONS;
             else allow += ", " + METHOD_OPTIONS;
-        
+
         resp.setHeader("Allow", allow);
     }
-    
-    
+
+
     /**
      * Called by the server (via the <code>service</code> method)
      * to allow a servlet to handle a TRACE request.
      *
      * A TRACE returns the headers sent with the TRACE
      * request to the client, so that they can be used in
-     * debugging. There's no need to override this method. 
+     * debugging. There's no need to override this method.
      *
      * @param req   the {@link HttpServletRequest} object that
      *                  contains the request the client made of
@@ -543,7 +543,7 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * @param resp  the {@link HttpServletResponse} object that
      *                  contains the response the servlet returns
-     *                  to the client                                
+     *                  to the client
      *
      * @exception IOException   if an input or output error occurs
      *                              while the servlet is handling the
@@ -552,42 +552,42 @@ public abstract class HttpServlet extends GenericServlet {
      * @exception ServletException  if the request for the
      *                                  TRACE cannot be handled
      */
-    protected void doTrace(HttpServletRequest req, HttpServletResponse resp) 
+    protected void doTrace(HttpServletRequest req, HttpServletResponse resp)
         throws ServletException, IOException
     {
-        
+
         int responseLength;
-        
+
         String CRLF = "\r\n";
         StringBuilder buffer = new StringBuilder("TRACE ").append(req.getRequestURI())
             .append(" ").append(req.getProtocol());
-        
+
         Enumeration<String> reqHeaderEnum = req.getHeaderNames();
-        
+
         while( reqHeaderEnum.hasMoreElements() ) {
             String headerName = reqHeaderEnum.nextElement();
             buffer.append(CRLF).append(headerName).append(": ")
-                .append(req.getHeader(headerName)); 
+                .append(req.getHeader(headerName));
         }
-        
+
         buffer.append(CRLF);
-        
+
         responseLength = buffer.length();
-        
+
         resp.setContentType("message/http");
         resp.setContentLength(responseLength);
         ServletOutputStream out = resp.getOutputStream();
-        out.print(buffer.toString());        
+        out.print(buffer.toString());
         out.close();
         return;
-    }                
+    }
 
 
     /**
      * Receives standard HTTP requests from the public
      * <code>service</code> method and dispatches
-     * them to the <code>do</code><i>XXX</i> methods defined in 
-     * this class. This method is an HTTP-specific version of the 
+     * them to the <code>do</code><i>Method</i> methods defined in
+     * this class. This method is an HTTP-specific version of the
      * {@link javax.servlet.Servlet#service} method. There's no
      * need to override this method.
      *
@@ -597,7 +597,7 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * @param resp  the {@link HttpServletResponse} object that
      *                  contains the response the servlet returns
-     *                  to the client                                
+     *                  to the client
      *
      * @exception IOException   if an input or output error occurs
      *                              while the servlet is handling the
@@ -605,7 +605,7 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * @exception ServletException  if the HTTP request
      *                                  cannot be handled
-     * 
+     *
      * @see javax.servlet.Servlet#service
      */
     protected void service(HttpServletRequest req, HttpServletResponse resp)
@@ -620,7 +620,13 @@ public abstract class HttpServlet extends GenericServlet {
                 // to go through further expensive logic
                 doGet(req, resp);
             } else {
-                long ifModifiedSince = req.getDateHeader(HEADER_IFMODSINCE);
+                long ifModifiedSince;
+                try {
+                    ifModifiedSince = req.getDateHeader(HEADER_IFMODSINCE);
+                } catch (IllegalArgumentException iae) {
+                    // Invalid date header - proceed as if none was set
+                    ifModifiedSince = -1;
+                }
                 if (ifModifiedSince < (lastModified / 1000 * 1000)) {
                     // If the servlet mod time is later, call doGet()
                     // Round down to the nearest second for a proper compare
@@ -639,19 +645,19 @@ public abstract class HttpServlet extends GenericServlet {
 
         } else if (method.equals(METHOD_POST)) {
             doPost(req, resp);
-            
+
         } else if (method.equals(METHOD_PUT)) {
-            doPut(req, resp);        
-            
+            doPut(req, resp);
+
         } else if (method.equals(METHOD_DELETE)) {
             doDelete(req, resp);
-            
+
         } else if (method.equals(METHOD_OPTIONS)) {
             doOptions(req,resp);
-            
+
         } else if (method.equals(METHOD_TRACE)) {
             doTrace(req,resp);
-            
+
         } else {
             //
             // Note that this means NO servlet supports whatever
@@ -662,7 +668,7 @@ public abstract class HttpServlet extends GenericServlet {
             Object[] errArgs = new Object[1];
             errArgs[0] = method;
             errMsg = MessageFormat.format(errMsg, errArgs);
-            
+
             resp.sendError(HttpServletResponse.SC_NOT_IMPLEMENTED, errMsg);
         }
     }
@@ -682,20 +688,20 @@ public abstract class HttpServlet extends GenericServlet {
         if (lastModified >= 0)
             resp.setDateHeader(HEADER_LASTMOD, lastModified);
     }
-   
-    
+
+
     /**
      * Dispatches client requests to the protected
      * <code>service</code> method. There's no need to
      * override this method.
-     * 
+     *
      * @param req   the {@link HttpServletRequest} object that
      *                  contains the request the client made of
      *                  the servlet
      *
      * @param res   the {@link HttpServletResponse} object that
      *                  contains the response the servlet returns
-     *                  to the client                                
+     *                  to the client
      *
      * @exception IOException   if an input or output error occurs
      *                              while the servlet is handling the
@@ -703,7 +709,7 @@ public abstract class HttpServlet extends GenericServlet {
      *
      * @exception ServletException  if the HTTP request cannot
      *                                  be handled
-     * 
+     *
      * @see javax.servlet.Servlet#service
      */
     @Override
@@ -712,7 +718,7 @@ public abstract class HttpServlet extends GenericServlet {
 
         HttpServletRequest  request;
         HttpServletResponse response;
-        
+
         try {
             request = (HttpServletRequest) req;
             response = (HttpServletResponse) res;
@@ -757,6 +763,36 @@ class NoBodyResponse extends HttpServletResponseWrapper {
         didSetContentLength = true;
     }
 
+    @Override
+    public void setHeader(String name, String value) {
+        super.setHeader(name, value);
+        checkHeader(name);
+    }
+
+    @Override
+    public void addHeader(String name, String value) {
+        super.addHeader(name, value);
+        checkHeader(name);
+    }
+
+    @Override
+    public void setIntHeader(String name, int value) {
+        super.setIntHeader(name, value);
+        checkHeader(name);
+    }
+
+    @Override
+    public void addIntHeader(String name, int value) {
+        super.addIntHeader(name, value);
+        checkHeader(name);
+    }
+
+    private void checkHeader(String name) {
+        if ("content-length".equalsIgnoreCase(name)) {
+            didSetContentLength = true;
+        }
+    }
+
     @Override
     public ServletOutputStream getOutputStream() throws IOException {
         return noBody;
@@ -779,7 +815,7 @@ class NoBodyResponse extends HttpServletResponseWrapper {
 /*
  * Servlet output stream that gobbles up all its data.
  */
- 
+
 // file private
 class NoBodyOutputStream extends ServletOutputStream {
 
@@ -806,17 +842,22 @@ class NoBodyOutputStream extends ServletOutputStream {
     }
 
     @Override
-    public void write(byte buf[], int offset, int len)
-        throws IOException
-    {
-        if (len >= 0) {
-            contentLength += len;
-        } else {
-            // XXX
-            // isn't this really an IllegalArgumentException?
-            
-            String msg = lStrings.getString("err.io.negativelength");
-            throw new IOException(msg);
+    public void write(byte buf[], int offset, int len) throws IOException {
+        if (buf == null) {
+            throw new NullPointerException(
+                    lStrings.getString("err.io.nullArray"));
         }
+
+        if (offset < 0 || len < 0 || offset+len > buf.length) {
+            String msg = lStrings.getString("err.io.indexOutOfBounds");
+            Object[] msgArgs = new Object[3];
+            msgArgs[0] = Integer.valueOf(offset);
+            msgArgs[1] = Integer.valueOf(len);
+            msgArgs[2] = Integer.valueOf(buf.length);
+            msg = MessageFormat.format(msg, msgArgs);
+            throw new IndexOutOfBoundsException(msg);
+        }
+
+        contentLength += len;
     }
 }
