@@ -44,6 +44,7 @@ import org.apache.catalina.AccessLog;
 import org.apache.catalina.Globals;
 import org.apache.catalina.LifecycleException;
 import org.apache.catalina.LifecycleState;
+import org.apache.catalina.Session;
 import org.apache.catalina.connector.Request;
 import org.apache.catalina.connector.Response;
 import org.apache.coyote.RequestInfo;
@@ -110,6 +111,8 @@ import org.apache.tomcat.util.buf.B2CConverter;
  * <li><code>%{xxx}c</code> for a specific cookie
  * <li><code>%{xxx}r</code> xxx is an attribute in the ServletRequest
  * <li><code>%{xxx}s</code> xxx is an attribute in the HttpSession
+ * <li><code>%{xxx}t</code> xxx is an enhanced SimpleDateFormat pattern
+ * (see Configuration Reference document for details on supported time patterns)
  * </ul>
  *
  * <p>
@@ -153,7 +156,7 @@ import org.apache.tomcat.util.buf.B2CConverter;
  * @author Takayuki Kaneko
  * @author Peter Rossbach
  *
- * @version $Id: AccessLogValve.java 1301255 2012-03-15 22:47:40Z markt $
+ * @version $Id: AccessLogValve.java 1416660 2012-12-03 20:47:57Z markt $
  */
 
 public class AccessLogValve extends ValveBase implements AccessLog {
@@ -210,6 +213,12 @@ public class AccessLogValve extends ValveBase implements AccessLog {
      */
     protected boolean rotatable = true;
 
+    /**
+     * Should we defer inclusion of the date stamp in the file
+     * name until rotate time? Default is false.
+     */
+    protected boolean renameOnRotate = false;
+
 
     /**
      * Buffered logging.
@@ -236,25 +245,6 @@ public class AccessLogValve extends ValveBase implements AccessLog {
     protected SimpleDateFormat fileDateFormatter = null;
 
 
-    /**
-     * The system timezone.
-     */
-    private static final TimeZone timezone;
-
-
-    /**
-     * The time zone offset relative to GMT in text form when daylight saving
-     * is not in operation.
-     */
-    private static final String timeZoneNoDST;
-
-
-    /**
-     * The time zone offset relative to GMT in text form when daylight saving
-     * is in operation.
-     */
-    private static final String timeZoneDST;
-
     /**
      * The size of our global date format cache
      */
@@ -298,28 +288,28 @@ public class AccessLogValve extends ValveBase implements AccessLog {
      * <p>This class uses a small thread local first level cache and a bigger
      * synchronized global second level cache.</p>
      */
-    private static class DateFormatCache {
+    protected static class DateFormatCache {
 
-        private class Cache {
+        protected class Cache {
 
             /* CLF log format */
-            private static final String cLFFormat = "dd/MMM/yyyy:HH:mm:ss";
+            private static final String cLFFormat = "dd/MMM/yyyy:HH:mm:ss Z";
 
             /* Second used to retrieve CLF format in most recent invocation */
-            private long previousSeconds = 0L;
+            private long previousSeconds = Long.MIN_VALUE;
             /* Value of CLF format retrieved in most recent invocation */
             private String previousFormat = "";
 
             /* First second contained in cache */
-            private long first = 0L;
+            private long first = Long.MIN_VALUE;
             /* Last second contained in cache */
-            private long last = 0L;
+            private long last = Long.MIN_VALUE;
             /* Index of "first" in the cyclic cache */
             private int offset = 0;
             /* Helper object to be able to call SimpleDateFormat.format(). */
             private final Date currentDate = new Date();
 
-            private final String cache[];
+            protected final String cache[];
             private SimpleDateFormat formatter;
             private boolean isCLF = false;
 
@@ -388,14 +378,16 @@ public class AccessLogValve extends ValveBase implements AccessLog {
                     for (int i = 1; i < seconds - last; i++) {
                         cache[(index + cacheSize - i) % cacheSize] = null;
                     }
-                    first = seconds - cacheSize;
+                    first = seconds - (cacheSize - 1);
                     last = seconds;
+                    offset = (index + 1) % cacheSize;
                 } else if (seconds < first) {
                     for (int i = 1; i < first - seconds; i++) {
                         cache[(index + i) % cacheSize] = null;
                     }
                     first = seconds;
-                    last = seconds + cacheSize;
+                    last = seconds + (cacheSize - 1);
+                    offset = index;
                 }
 
                 /* Last step: format new timestamp either using
@@ -411,8 +403,6 @@ public class AccessLogValve extends ValveBase implements AccessLog {
                         StringBuilder current = new StringBuilder(32);
                         current.append('[');
                         current.append(previousFormat);
-                        current.append(' ');
-                        current.append(getTimeZone(currentDate));
                         current.append(']');
                         previousFormat = current.toString();
                     }
@@ -427,10 +417,10 @@ public class AccessLogValve extends ValveBase implements AccessLog {
 
         private final Locale cacheDefaultLocale;
         private final DateFormatCache parent;
-        private final Cache cLFCache;
+        protected final Cache cLFCache;
         private final HashMap<String, Cache> formatCache = new HashMap<String, Cache>();
 
-        private DateFormatCache(int size, Locale loc, DateFormatCache parent) {
+        protected DateFormatCache(int size, Locale loc, DateFormatCache parent) {
             cacheSize = size;
             cacheDefaultLocale = loc;
             this.parent = parent;
@@ -494,8 +484,7 @@ public class AccessLogValve extends ValveBase implements AccessLog {
      * The system time when we last updated the Date that this valve
      * uses for log lines.
      */
-    private static final ThreadLocal<Date> localDate =
-            new ThreadLocal<Date>() {
+    private static final ThreadLocal<Date> localDate = new ThreadLocal<Date>() {
         @Override
         protected Date initialValue() {
             return new Date();
@@ -505,7 +494,7 @@ public class AccessLogValve extends ValveBase implements AccessLog {
     /**
      * The list of our format types.
      */
-    private static enum formatType {
+    private static enum FormatType {
         CLF, SEC, MSEC, MSEC_FRAC, SDF
     }
 
@@ -720,6 +709,26 @@ public class AccessLogValve extends ValveBase implements AccessLog {
     }
 
 
+    /**
+     * Should we defer inclusion of the date stamp in the file
+     * name until rotate time
+     */
+    public boolean isRenameOnRotate() {
+        return renameOnRotate;
+    }
+
+
+    /**
+     * Set the value if we should defer inclusion of the date
+     * stamp in the file name until rotate time
+     *
+     * @param renameOnRotate true if defer inclusion of date stamp
+     */
+    public void setRenameOnRotate(boolean renameOnRotate) {
+        this.renameOnRotate = renameOnRotate;
+    }
+
+
     /**
      * Is the logging buffered
      */
@@ -970,7 +979,7 @@ public class AccessLogValve extends ValveBase implements AccessLog {
 
         if (currentLogFile != null) {
             File holder = currentLogFile;
-            close();
+            close(false);
             try {
                 holder.renameTo(new File(newFileName));
             } catch (Throwable e) {
@@ -994,14 +1003,87 @@ public class AccessLogValve extends ValveBase implements AccessLog {
 
 
     /**
-     * Close the currently open log file (if any)
+     * Create a File object based on the current log file name.
+     * Directories are created as needed but the underlying file
+     * is not created or opened.
+     *
+     * @param useDateStamp include the timestamp in the file name.
+     * @return the log file object
      */
-    private synchronized void close() {
+    private File getLogFile(boolean useDateStamp) {
+
+        // Create the directory if necessary
+        File dir = new File(directory);
+        if (!dir.isAbsolute()) {
+            dir = new File(System.getProperty(Globals.CATALINA_BASE_PROP), directory);
+        }
+        if (!dir.mkdirs() && !dir.isDirectory()) {
+            log.error(sm.getString("accessLogValve.openDirFail", dir));
+        }
+
+        // Calculate the current log file name
+        File pathname;
+        if (useDateStamp) {
+            pathname = new File(dir.getAbsoluteFile(), prefix + dateStamp
+                    + suffix);
+        } else {
+            pathname = new File(dir.getAbsoluteFile(), prefix + suffix);
+        }
+        File parent = pathname.getParentFile();
+        if (!parent.mkdirs() && !parent.isDirectory()) {
+            log.error(sm.getString("accessLogValve.openDirFail", parent));
+        }
+        return pathname;
+    }
+
+    /**
+     * Move a current but rotated log file back to the unrotated
+     * one. Needed if date stamp inclusion is deferred to rotation
+     * time.
+     */
+    private void restore() {
+        File newLogFile = getLogFile(false);
+        File rotatedLogFile = getLogFile(true);
+        if (rotatedLogFile.exists() && !newLogFile.exists() &&
+            !rotatedLogFile.equals(newLogFile)) {
+            try {
+                if (!rotatedLogFile.renameTo(newLogFile)) {
+                    log.error(sm.getString("accessLogValve.renameFail", rotatedLogFile, newLogFile));
+                }
+            } catch (Throwable e) {
+                ExceptionUtils.handleThrowable(e);
+                log.error(sm.getString("accessLogValve.renameFail", rotatedLogFile, newLogFile), e);
+            }
+        }
+    }
+
+
+    /**
+     * Close the currently open log file (if any)
+     *
+     * @param rename Rename file to final name after closing
+     */
+    private synchronized void close(boolean rename) {
         if (writer == null) {
             return;
         }
         writer.flush();
         writer.close();
+        if (rename && renameOnRotate) {
+            File newLogFile = getLogFile(true);
+            if (!newLogFile.exists()) {
+                try {
+                    if (!currentLogFile.renameTo(newLogFile)) {
+                        log.error(sm.getString("accessLogValve.renameFail", currentLogFile, newLogFile));
+                    }
+                } catch (Throwable e) {
+                    ExceptionUtils.handleThrowable(e);
+                    log.error(sm.getString("accessLogValve.renameFail", currentLogFile, newLogFile), e);
+                }
+            } else {
+                log.error(sm.getString("accessLogValve.alreadyExists", currentLogFile, newLogFile));
+            }
+        }
         writer = null;
         dateStamp = "";
         currentLogFile = null;
@@ -1029,7 +1111,7 @@ public class AccessLogValve extends ValveBase implements AccessLog {
 
                         // If the date has changed, switch log files
                         if (!dateStamp.equals(tsDate)) {
-                            close();
+                            close(true);
                             dateStamp = tsDate;
                             open();
                         }
@@ -1043,7 +1125,7 @@ public class AccessLogValve extends ValveBase implements AccessLog {
             synchronized (this) {
                 if (currentLogFile != null && !currentLogFile.exists()) {
                     try {
-                        close();
+                        close(false);
                     } catch (Throwable e) {
                         ExceptionUtils.handleThrowable(e);
                         log.info(sm.getString("accessLogValve.closeFail"), e);
@@ -1075,28 +1157,9 @@ public class AccessLogValve extends ValveBase implements AccessLog {
      * Open the new log file for the date specified by <code>dateStamp</code>.
      */
     protected synchronized void open() {
-        // Create the directory if necessary
-        File dir = new File(directory);
-        if (!dir.isAbsolute()) {
-            dir = new File(System.getProperty(Globals.CATALINA_BASE_PROP), directory);
-        }
-        if (!dir.mkdirs() && !dir.isDirectory()) {
-            log.error(sm.getString("accessLogValve.openDirFail", dir));
-        }
-
         // Open the current log file
-        File pathname;
         // If no rotate - no need for dateStamp in fileName
-        if (rotatable) {
-            pathname = new File(dir.getAbsoluteFile(), prefix + dateStamp
-                    + suffix);
-        } else {
-            pathname = new File(dir.getAbsoluteFile(), prefix + suffix);
-        }
-        File parent = pathname.getParentFile();
-        if (!parent.mkdirs() && !parent.isDirectory()) {
-            log.error(sm.getString("accessLogValve.openDirFail", parent));
-        }
+        File pathname = getLogFile(rotatable && !renameOnRotate);
 
         Charset charset = null;
         if (encoding != null) {
@@ -1125,11 +1188,8 @@ public class AccessLogValve extends ValveBase implements AccessLog {
     }
 
     /**
-     * This method returns a Date object that is accurate to within one second.
-     * If a thread calls this method to get a Date and it's been less than 1
-     * second since a new Date was created, this method simply gives out the
-     * same Date again so that the system doesn't spend time creating Date
-     * objects unnecessarily.
+     * This method returns a ThreadLocal Date object that is set to the
+     * specified time. This saves creating a new Date object for every request.
      *
      * @return Date
      */
@@ -1140,40 +1200,6 @@ public class AccessLogValve extends ValveBase implements AccessLog {
     }
 
 
-    private static String getTimeZone(Date date) {
-        if (timezone.inDaylightTime(date)) {
-            return timeZoneDST;
-        } else {
-            return timeZoneNoDST;
-        }
-    }
-
-
-    private static String calculateTimeZoneOffset(long offset) {
-        StringBuilder tz = new StringBuilder();
-        if ((offset < 0)) {
-            tz.append("-");
-            offset = -offset;
-        } else {
-            tz.append("+");
-        }
-
-        long hourOffset = offset / (1000 * 60 * 60);
-        long minuteOffset = (offset / (1000 * 60)) % 60;
-
-        if (hourOffset < 10) {
-            tz.append("0");
-        }
-        tz.append(hourOffset);
-
-        if (minuteOffset < 10) {
-            tz.append("0");
-        }
-        tz.append(minuteOffset);
-
-        return tz.toString();
-    }
-
     /**
      * Find a locale by name
      */
@@ -1191,14 +1217,6 @@ public class AccessLogValve extends ValveBase implements AccessLog {
         return fallback;
     }
 
-    static {
-        // Initialize the timeZone
-        timezone = TimeZone.getDefault();
-        timeZoneNoDST = calculateTimeZoneOffset(timezone.getRawOffset());
-        int offset = timezone.getDSTSavings();
-        timeZoneDST = calculateTimeZoneOffset(timezone.getRawOffset() + offset);
-    }
-
 
     /**
      * Start this component and implement the requirements
@@ -1217,8 +1235,11 @@ public class AccessLogValve extends ValveBase implements AccessLog {
             setFileDateFormat(format);
         }
         fileDateFormatter = new SimpleDateFormat(format, Locale.US);
-        fileDateFormatter.setTimeZone(timezone);
+        fileDateFormatter.setTimeZone(TimeZone.getDefault());
         dateStamp = fileDateFormatter.format(new Date(System.currentTimeMillis()));
+        if (rotatable && renameOnRotate) {
+            restore();
+        }
         open();
 
         setState(LifecycleState.STARTING);
@@ -1236,7 +1257,7 @@ public class AccessLogValve extends ValveBase implements AccessLog {
     protected synchronized void stopInternal() throws LifecycleException {
 
         setState(LifecycleState.STOPPING);
-        close();
+        close(false);
     }
 
     /**
@@ -1432,7 +1453,7 @@ public class AccessLogValve extends ValveBase implements AccessLog {
         /* Whether to use begin of request or end of response as the timestamp */
         private boolean usesBegin = false;
         /* The format type */
-        private formatType type = formatType.CLF;
+        private FormatType type = FormatType.CLF;
         /* Whether we need to postprocess by adding milliseconds */
         private boolean usesMsecs = false;
 
@@ -1484,15 +1505,15 @@ public class AccessLogValve extends ValveBase implements AccessLog {
                     format = format.substring(4);
                 }
                 if (format.length() == 0) {
-                    type = formatType.CLF;
+                    type = FormatType.CLF;
                 } else if (format.equals(secFormat)) {
-                    type = formatType.SEC;
+                    type = FormatType.SEC;
                 } else if (format.equals(msecFormat)) {
-                    type = formatType.MSEC;
+                    type = FormatType.MSEC;
                 } else if (format.equals(msecFractionFormat)) {
-                    type = formatType.MSEC_FRAC;
+                    type = FormatType.MSEC_FRAC;
                 } else {
-                    type = formatType.SDF;
+                    type = FormatType.SDF;
                     format = tidyFormat(format);
                 }
             }
@@ -1726,15 +1747,15 @@ public class AccessLogValve extends ValveBase implements AccessLog {
         @Override
         public void addElement(StringBuilder buf, Date date, Request request,
                 Response response, long time) {
-            if (request != null) {
-                if (request.getSession(false) != null) {
-                    buf.append(request.getSessionInternal(false)
-                            .getIdInternal());
-                } else {
-                    buf.append('-');
-                }
-            } else {
+            if (request == null) {
                 buf.append('-');
+            } else {
+                Session session = request.getSessionInternal(false);
+                if (session == null) {
+                    buf.append('-');
+                } else {
+                    buf.append(session.getIdInternal());
+                }
             }
         }
     }
