@@ -26,11 +26,13 @@ import java.util.Arrays;
 import java.util.Date;
 import java.util.Enumeration;
 import java.util.Hashtable;
+import java.util.List;
 import java.util.zip.ZipEntry;
 import java.util.zip.ZipException;
 import java.util.zip.ZipFile;
 
 import javax.naming.CompositeName;
+import javax.naming.InvalidNameException;
 import javax.naming.Name;
 import javax.naming.NamingEnumeration;
 import javax.naming.NamingException;
@@ -39,21 +41,20 @@ import javax.naming.directory.Attributes;
 import javax.naming.directory.DirContext;
 import javax.naming.directory.ModificationItem;
 import javax.naming.directory.SearchControls;
+import javax.naming.directory.SearchResult;
 
-import org.apache.naming.NamingContextBindingsEnumeration;
-import org.apache.naming.NamingContextEnumeration;
 import org.apache.naming.NamingEntry;
 
 /**
  * WAR Directory Context implementation.
  *
  * @author Remy Maucherat
- * @version $Revision: 812112 $ $Date: 2009-09-07 14:37:06 +0200 (Mo, 07. Sep 2009) $
+ * @version $Id: WARDirContext.java 1392099 2012-09-30 19:48:36Z markt $
  */
 
 public class WARDirContext extends BaseDirContext {
 
-    private static org.apache.juli.logging.Log log=
+    private static final org.apache.juli.logging.Log log=
         org.apache.juli.logging.LogFactory.getLog( WARDirContext.class );
     
     // ----------------------------------------------------------- Constructors
@@ -70,13 +71,14 @@ public class WARDirContext extends BaseDirContext {
     /**
      * Builds a WAR directory context using the given environment.
      */
-    public WARDirContext(Hashtable env) {
+    public WARDirContext(Hashtable<String,Object> env) {
         super(env);
     }
 
 
     /**
-     * Constructor used for returning fake subcontexts.
+     * Constructor used for returning fake sub-contexts or for accessing
+     * META-INF/resources locations in bundled JAR files.
      */
     protected WARDirContext(ZipFile base, Entry entries) {
         this.base = base;
@@ -112,28 +114,29 @@ public class WARDirContext extends BaseDirContext {
      * @exception IllegalArgumentException if this would create a
      *  malformed URL
      */
+    @Override
     public void setDocBase(String docBase) {
 
-	// Validate the format of the proposed document root
-	if (docBase == null)
-	    throw new IllegalArgumentException
-		(sm.getString("resources.null"));
-	if (!(docBase.endsWith(".war")))
-	    throw new IllegalArgumentException
-		(sm.getString("warResources.notWar"));
+        // Validate the format of the proposed document root
+        if (docBase == null)
+            throw new IllegalArgumentException
+                (sm.getString("resources.null"));
+        if (!(docBase.endsWith(".war")))
+            throw new IllegalArgumentException
+                (sm.getString("warResources.notWar"));
 
-	// Calculate a File object referencing this document base directory
-	File base = new File(docBase);
+        // Calculate a File object referencing this document base directory
+        File base = new File(docBase);
 
-	// Validate that the document base is an existing directory
-	if (!base.exists() || !base.canRead() || base.isDirectory())
-	    throw new IllegalArgumentException
-		(sm.getString("warResources.invalidWar", docBase));
+        // Validate that the document base is an existing directory
+        if (!base.exists() || !base.canRead() || base.isDirectory())
+            throw new IllegalArgumentException
+                (sm.getString("warResources.invalidWar", docBase));
         try {
             this.base = new ZipFile(base);
         } catch (Exception e) {
-	    throw new IllegalArgumentException
-		(sm.getString("warResources.invalidWar", e.getMessage()));
+            throw new IllegalArgumentException
+                (sm.getString("warResources.invalidWar", e.getMessage()));
         }
         super.setDocBase(docBase);
 
@@ -148,6 +151,7 @@ public class WARDirContext extends BaseDirContext {
     /**
      * Release any resources allocated for this directory context.
      */
+    @Override
     public void release() {
 
         entries = null;
@@ -164,6 +168,17 @@ public class WARDirContext extends BaseDirContext {
 
     }
 
+    /**
+     * Return the real path for a given virtual path, if possible; otherwise
+     * return <code>null</code>.
+     *
+     * @param path The path to the desired resource
+     */
+    @Override
+    protected String doGetRealPath(String path) {
+        return null;
+    }
+
 
     // -------------------------------------------------------- Context Methods
 
@@ -171,34 +186,26 @@ public class WARDirContext extends BaseDirContext {
     /**
      * Retrieves the named object.
      * 
-     * @param name the name of the object to look up
+     * @param strName the name of the object to look up
      * @return the object bound to name
-     * @exception NamingException if a naming exception is encountered
      */
-    public Object lookup(String name)
-        throws NamingException {
-        return lookup(new CompositeName(name));
-    }
+    @Override
+    protected Object doLookup(String strName) {
 
+        Name name;
+        try {
+            name = getEscapedJndiName(strName);
+        } catch (InvalidNameException e) {
+            log.info(sm.getString("resources.invalidName", strName), e);
+            return null;
+        }
 
-    /**
-     * Retrieves the named object. If name is empty, returns a new instance 
-     * of this context (which represents the same naming context as this 
-     * context, but its environment may be modified independently and it may 
-     * be accessed concurrently).
-     * 
-     * @param name the name of the object to look up
-     * @return the object bound to name
-     * @exception NamingException if a naming exception is encountered
-     */
-    public Object lookup(Name name)
-        throws NamingException {
         if (name.isEmpty())
             return this;
         Entry entry = treeLookup(name);
         if (entry == null)
-            throw new NamingException
-                (sm.getString("resources.notFound", name));
+            return null;
+            
         ZipEntry zipEntry = entry.getEntry();
         if (zipEntry.isDirectory())
             return new WARDirContext(base, entry);
@@ -207,6 +214,19 @@ public class WARDirContext extends BaseDirContext {
     }
 
 
+    /**
+     * JNDI treats ' and " as reserved characters therefore they need to be
+     * escaped as part of converting file names to JNDI names. Note that while
+     * ' can be used in Windows and Unix file names, " is only valid on Unix.
+     * This method assumes that the string is currently unquoted.
+     * 
+     * @return  A valid JNDI name
+     * @throws InvalidNameException 
+     */
+    private Name getEscapedJndiName(String name) throws InvalidNameException {
+        return new CompositeName(name.replace("'", "\\'").replace("\"", ""));
+    }
+
     /**
      * Unbinds the named object. Removes the terminal atomic name in name 
      * from the target context--that named by all but the terminal atomic 
@@ -221,6 +241,7 @@ public class WARDirContext extends BaseDirContext {
      * exist
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public void unbind(String name)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -235,59 +256,17 @@ public class WARDirContext extends BaseDirContext {
      * 
      * @param oldName the name of the existing binding; may not be empty
      * @param newName the name of the new binding; may not be empty
-     * @exception NameAlreadyBoundException if newName is already bound
+     * @exception javax.naming.NameAlreadyBoundException if newName is already
+     * bound
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public void rename(String oldName, String newName)
         throws NamingException {
         throw new OperationNotSupportedException();
     }
 
 
-    /**
-     * Enumerates the names bound in the named context, along with the class 
-     * names of objects bound to them. The contents of any subcontexts are 
-     * not included.
-     * <p>
-     * If a binding is added to or removed from this context, its effect on 
-     * an enumeration previously returned is undefined.
-     * 
-     * @param name the name of the context to list
-     * @return an enumeration of the names and class names of the bindings in 
-     * this context. Each element of the enumeration is of type NameClassPair.
-     * @exception NamingException if a naming exception is encountered
-     */
-    public NamingEnumeration list(String name)
-        throws NamingException {
-        return list(new CompositeName(name));
-    }
-
-
-    /**
-     * Enumerates the names bound in the named context, along with the class 
-     * names of objects bound to them. The contents of any subcontexts are 
-     * not included.
-     * <p>
-     * If a binding is added to or removed from this context, its effect on 
-     * an enumeration previously returned is undefined.
-     * 
-     * @param name the name of the context to list
-     * @return an enumeration of the names and class names of the bindings in 
-     * this context. Each element of the enumeration is of type NameClassPair.
-     * @exception NamingException if a naming exception is encountered
-     */
-    public NamingEnumeration list(Name name)
-        throws NamingException {
-        if (name.isEmpty())
-            return new NamingContextEnumeration(list(entries).iterator());
-        Entry entry = treeLookup(name);
-        if (entry == null)
-            throw new NamingException
-                (sm.getString("resources.notFound", name));
-        return new NamingContextEnumeration(list(entry).iterator());
-    }
-
-
     /**
      * Enumerates the names bound in the named context, along with the 
      * objects bound to them. The contents of any subcontexts are not 
@@ -296,41 +275,25 @@ public class WARDirContext extends BaseDirContext {
      * If a binding is added to or removed from this context, its effect on 
      * an enumeration previously returned is undefined.
      * 
-     * @param name the name of the context to list
+     * @param strName the name of the context to list
      * @return an enumeration of the bindings in this context. 
      * Each element of the enumeration is of type Binding.
      * @exception NamingException if a naming exception is encountered
      */
-    public NamingEnumeration listBindings(String name)
+    @Override
+    protected List<NamingEntry> doListBindings(String strName)
         throws NamingException {
-        return listBindings(new CompositeName(name));
-    }
+        
+        Name name = getEscapedJndiName(strName);
 
-
-    /**
-     * Enumerates the names bound in the named context, along with the 
-     * objects bound to them. The contents of any subcontexts are not 
-     * included.
-     * <p>
-     * If a binding is added to or removed from this context, its effect on 
-     * an enumeration previously returned is undefined.
-     * 
-     * @param name the name of the context to list
-     * @return an enumeration of the bindings in this context. 
-     * Each element of the enumeration is of type Binding.
-     * @exception NamingException if a naming exception is encountered
-     */
-    public NamingEnumeration listBindings(Name name)
-        throws NamingException {
         if (name.isEmpty())
-            return new NamingContextBindingsEnumeration(list(entries).iterator(),
-                    this);
+            return list(entries);
+
         Entry entry = treeLookup(name);
         if (entry == null)
-            throw new NamingException
-                (sm.getString("resources.notFound", name));
-        return new NamingContextBindingsEnumeration(list(entry).iterator(),
-                this);
+            return null;
+        
+        return list(entry);
     }
 
 
@@ -356,9 +319,10 @@ public class WARDirContext extends BaseDirContext {
      * @param name the name of the context to be destroyed; may not be empty
      * @exception NameNotFoundException if an intermediate context does not 
      * exist
-     * @exception NotContextException if the name is bound but does not name 
-     * a context, or does not name a context of the appropriate type
+     * @exception javax.naming.NotContextException if the name is bound but does
+     * not name a context, or does not name a context of the appropriate type
      */
+    @Override
     public void destroySubcontext(String name)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -375,6 +339,7 @@ public class WARDirContext extends BaseDirContext {
      * (if any).
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public Object lookupLink(String name)
         throws NamingException {
         // Note : Links are not supported
@@ -399,6 +364,7 @@ public class WARDirContext extends BaseDirContext {
      * not have the notion of a full name
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public String getNameInNamespace()
         throws NamingException {
         return docBase;
@@ -420,9 +386,10 @@ public class WARDirContext extends BaseDirContext {
      * indicates that none should be retrieved
      * @exception NamingException if a naming exception is encountered
      */
-    public Attributes getAttributes(String name, String[] attrIds)
+    @Override
+    protected Attributes doGetAttributes(String name, String[] attrIds)
         throws NamingException {
-        return getAttributes(new CompositeName(name), attrIds);
+        return getAttributes(getEscapedJndiName(name), attrIds);
     }
 
 
@@ -434,6 +401,7 @@ public class WARDirContext extends BaseDirContext {
      * @param name the name of the object from which to retrieve attributes
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public Attributes getAttributes(Name name, String[] attrIds)
         throws NamingException {
         
@@ -443,8 +411,7 @@ public class WARDirContext extends BaseDirContext {
         else
             entry = treeLookup(name);
         if (entry == null)
-            throw new NamingException
-                (sm.getString("resources.notFound", name));
+            return null;
         
         ZipEntry zipEntry = entry.getEntry();
 
@@ -453,6 +420,8 @@ public class WARDirContext extends BaseDirContext {
         attrs.setName(entry.getName());
         if (!zipEntry.isDirectory())
             attrs.setResourceType("");
+        else
+            attrs.setCollection(true);
         attrs.setContentLength(zipEntry.getSize());
         attrs.setLastModified(zipEntry.getTime());
         
@@ -471,10 +440,11 @@ public class WARDirContext extends BaseDirContext {
      * REPLACE_ATTRIBUTE, REMOVE_ATTRIBUTE
      * @param attrs the attributes to be used for the modification; may not 
      * be null
-     * @exception AttributeModificationException if the modification cannot be
-     * completed successfully
+     * @exception javax.naming.directory.AttributeModificationException if the
+     * modification cannot be completed successfully
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public void modifyAttributes(String name, int mod_op, Attributes attrs)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -491,10 +461,11 @@ public class WARDirContext extends BaseDirContext {
      * @param name the name of the object whose attributes will be updated
      * @param mods an ordered sequence of modifications to be performed; may 
      * not be null
-     * @exception AttributeModificationException if the modification cannot be
-     * completed successfully
+     * @exception javax.naming.directory.AttributeModificationException if the
+     * modification cannot be completed successfully
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public void modifyAttributes(String name, ModificationItem[] mods)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -511,11 +482,13 @@ public class WARDirContext extends BaseDirContext {
      * @param name the name to bind; may not be empty
      * @param obj the object to bind; possibly null
      * @param attrs the attributes to associate with the binding
-     * @exception NameAlreadyBoundException if name is already bound
-     * @exception InvalidAttributesException if some "mandatory" attributes 
-     * of the binding are not supplied
+     * @exception javax.naming.NameAlreadyBoundException if name is already
+     * bound
+     * @exception javax.naming.directory.InvalidAttributesException if some
+     * "mandatory" attributes of the binding are not supplied
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public void bind(String name, Object obj, Attributes attrs)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -536,10 +509,11 @@ public class WARDirContext extends BaseDirContext {
      * @param name the name to bind; may not be empty
      * @param obj the object to bind; possibly null
      * @param attrs the attributes to associate with the binding
-     * @exception InvalidAttributesException if some "mandatory" attributes 
-     * of the binding are not supplied
+     * @exception javax.naming.directory.InvalidAttributesException if some
+     * "mandatory" attributes of the binding are not supplied
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public void rebind(String name, Object obj, Attributes attrs)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -558,11 +532,13 @@ public class WARDirContext extends BaseDirContext {
      * @param name the name of the context to create; may not be empty
      * @param attrs the attributes to associate with the newly created context
      * @return the newly created context
-     * @exception NameAlreadyBoundException if the name is already bound
-     * @exception InvalidAttributesException if attrs does not contain all 
-     * the mandatory attributes required for creation
+     * @exception javax.naming.NameAlreadyBoundException if name is already
+     * bound
+     * @exception javax.naming.directory.InvalidAttributesException if attrs
+     * does not contain all the mandatory attributes required for creation
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public DirContext createSubcontext(String name, Attributes attrs)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -582,6 +558,7 @@ public class WARDirContext extends BaseDirContext {
      * @exception OperationNotSupportedException if schema not supported
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public DirContext getSchema(String name)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -599,6 +576,7 @@ public class WARDirContext extends BaseDirContext {
      * @exception OperationNotSupportedException if schema not supported
      * @exception NamingException if a naming exception is encountered
      */
+    @Override
     public DirContext getSchemaClassDefinition(String name)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -622,8 +600,9 @@ public class WARDirContext extends BaseDirContext {
      * context named by name.
      * @exception NamingException if a naming exception is encountered
      */
-    public NamingEnumeration search(String name, Attributes matchingAttributes,
-                                    String[] attributesToReturn)
+    @Override
+    public NamingEnumeration<SearchResult> search(String name,
+            Attributes matchingAttributes, String[] attributesToReturn)
         throws NamingException {
         throw new OperationNotSupportedException();
     }
@@ -644,8 +623,9 @@ public class WARDirContext extends BaseDirContext {
      * context named by name.
      * @exception NamingException if a naming exception is encountered
      */
-    public NamingEnumeration search(String name, Attributes matchingAttributes)
-        throws NamingException {
+    @Override
+    public NamingEnumeration<SearchResult> search(String name,
+            Attributes matchingAttributes) throws NamingException {
         throw new OperationNotSupportedException();
     }
 
@@ -663,13 +643,15 @@ public class WARDirContext extends BaseDirContext {
      * (new SearchControls())).
      * @return an enumeration of SearchResults of the objects that satisfy 
      * the filter; never null
-     * @exception InvalidSearchFilterException if the search filter specified 
-     * is not supported or understood by the underlying directory
-     * @exception InvalidSearchControlsException if the search controls 
-     * contain invalid settings
+     * @exception javax.naming.directory.InvalidSearchFilterException if the
+     * search filter specified is not supported or understood by the underlying
+     * directory
+     * @exception javax.naming.directory.InvalidSearchControlsException if the
+     * search controls contain invalid settings
      * @exception NamingException if a naming exception is encountered
      */
-    public NamingEnumeration search(String name, String filter, 
+    @Override
+    public NamingEnumeration<SearchResult> search(String name, String filter, 
                                     SearchControls cons)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -690,17 +672,18 @@ public class WARDirContext extends BaseDirContext {
      * occurrence of "{i}". If null, equivalent to an empty array.
      * @param cons the search controls that control the search. If null, the 
      * default search controls are used (equivalent to (new SearchControls())).
-     * @return an enumeration of SearchResults of the objects that satisy the 
+     * @return an enumeration of SearchResults of the objects that satisfy the 
      * filter; never null
      * @exception ArrayIndexOutOfBoundsException if filterExpr contains {i} 
      * expressions where i is outside the bounds of the array filterArgs
-     * @exception InvalidSearchControlsException if cons contains invalid 
-     * settings
-     * @exception InvalidSearchFilterException if filterExpr with filterArgs 
-     * represents an invalid search filter
+     * @exception javax.naming.directory.InvalidSearchControlsException if cons
+     * contains invalid settings
+     * @exception javax.naming.directory.InvalidSearchFilterException if
+     * filterExpr with filterArgs represents an invalid search filter
      * @exception NamingException if a naming exception is encountered
      */
-    public NamingEnumeration search(String name, String filterExpr, 
+    @Override
+    public NamingEnumeration<SearchResult> search(String name, String filterExpr, 
                                     Object[] filterArgs, SearchControls cons)
         throws NamingException {
         throw new OperationNotSupportedException();
@@ -731,12 +714,12 @@ public class WARDirContext extends BaseDirContext {
 
         try {
 
-            Enumeration entryList = base.entries();
+            Enumeration<? extends ZipEntry> entryList = base.entries();
             entries = new Entry("/", new ZipEntry("/"));
             
             while (entryList.hasMoreElements()) {
                 
-                ZipEntry entry = (ZipEntry) entryList.nextElement();
+                ZipEntry entry = entryList.nextElement();
                 String name = normalize(entry);
                 int pos = name.lastIndexOf('/');
                 // Check that parent entries exist and, if not, create them.
@@ -745,8 +728,8 @@ public class WARDirContext extends BaseDirContext {
                 int currentPos = -1;
                 int lastPos = 0;
                 while ((currentPos = name.indexOf('/', lastPos)) != -1) {
-                    Name parentName = new CompositeName(name.substring(0, lastPos));
-                    Name childName = new CompositeName(name.substring(0, currentPos));
+                    Name parentName = getEscapedJndiName(name.substring(0, lastPos));
+                    Name childName = getEscapedJndiName(name.substring(0, currentPos));
                     String entryName = name.substring(lastPos, currentPos);
                     // Parent should have been created in last cycle through
                     // this loop
@@ -766,7 +749,7 @@ public class WARDirContext extends BaseDirContext {
                     lastPos = currentPos + 1;
                 }
                 String entryName = name.substring(pos + 1, name.length());
-                Name compositeName = new CompositeName(name.substring(0, pos));
+                Name compositeName = getEscapedJndiName(name.substring(0, pos));
                 Entry parent = treeLookup(compositeName);
                 Entry child = new Entry(entryName, entry);
                 if (parent != null)
@@ -775,6 +758,7 @@ public class WARDirContext extends BaseDirContext {
             }
 
         } catch (Exception e) {
+            // Ignore
         }
 
     }
@@ -801,9 +785,9 @@ public class WARDirContext extends BaseDirContext {
     /**
      * List children as objects.
      */
-    protected ArrayList list(Entry entry) {
+    protected ArrayList<NamingEntry> list(Entry entry) {
         
-        ArrayList entries = new ArrayList();
+        ArrayList<NamingEntry> entries = new ArrayList<NamingEntry>();
         Entry[] children = entry.getChildren();
         Arrays.sort(children);
         NamingEntry namingEntry = null;
@@ -832,7 +816,7 @@ public class WARDirContext extends BaseDirContext {
     /**
      * Entries structure.
      */
-    protected class Entry implements Comparable {
+    protected static class Entry implements Comparable<Object> {
 
 
         // -------------------------------------------------------- Constructor
@@ -859,10 +843,23 @@ public class WARDirContext extends BaseDirContext {
         // ----------------------------------------------------- Public Methods
         
         
+        @Override
         public int compareTo(Object o) {
             if (!(o instanceof Entry))
                 return (+1);
-            return (name.compareTo(((Entry) o).getName()));
+            return name.compareTo(((Entry) o).getName());
+        }
+
+        @Override
+        public boolean equals(Object o) {
+            if (!(o instanceof Entry))
+                return false;
+            return name.equals(((Entry) o).getName());
+        }
+        
+        @Override
+        public int hashCode() {
+            return name.hashCode();
         }
 
         public ZipEntry getEntry() {
@@ -934,14 +931,17 @@ public class WARDirContext extends BaseDirContext {
          * 
          * @return InputStream
          */
+        @Override
         public InputStream streamContent()
             throws IOException {
             try {
                 if (binaryContent == null) {
-                    inputStream = base.getInputStream(entry);
+                    InputStream is = base.getInputStream(entry);
+                    inputStream = is;
+                    return is;
                 }
             } catch (ZipException e) {
-                throw new IOException(e.getMessage());
+                throw new IOException(e.getMessage(), e);
             }
             return super.streamContent();
         }
